/*******************************************************************************
* FilterMidi: A filter to obtain the copyright, lyrics trackname and any text
*             in a midi. The biggest benefit is the lyric support.. 
*
* Copyright 2004 by Andrew Luecke <auzy@users.sourceforge.net>
*
* I used  the FilterPdf code as reference by:  
*     Christopher Orr <dashboard@protactin.co.uk>
*
* The variable integer Code was based on: 
*     http://www.borg.com/~jglatt/tech/midifile/vari.htm
*
* For information on the midi file format, I used these sites: 
*  http://www.csw2.co.uk/tech/midi2.htm
*  http://www.sonicspot.com/guide/midifiles.html
*  http://www.borg.com/~jglatt/tech/midifile.htm
*  http://ourworld.compuserve.com/homepages/mark_clay/midi.htm
*  http://ourworld.compuserve.com/homepages/mark_clay/midifile.htm
*  http://crystal.apana.org.au/ghansper/midi_introduction/midi_file_format.html 
*******************************************************************************/

using System;
using System.IO;
using System.Security.Permissions;
using System.Diagnostics;

namespace Beagle.Filters {

	class FilterMidi : Beagle.Daemon.Filter {
		
		FileStream fs;
		string lyrics;
		string copyright;	
		string randomText;
		string trackName;

		/**********************************************************
		 * Function: FilterMidi()
		 * Purpose: To clear the properties
		 **********************************************************/
		public FilterMidi() 
		{
			lyrics 		= "";
			copyright  	= "";	
			randomText 	= "";
			trackName 	= "";
			AddSupportedMimeType ("audio/midi");
			AddSupportedExtension (".mid");
		}


		/***********************************************************
		 * Function: uint32ConvEndian()
		 * Purpose: To convert between endians into a uint32
		 ***********************************************************/
		UInt32 UInt32ConvEndian(byte[] data) 
		{
			byte [] goodBytes = new byte [data.Length];
		
			/* Swap the bits here */
			for (int cnt = 0; cnt < data.Length; cnt ++) 
			{
				goodBytes [cnt] = data [data.Length - 1 - cnt];
			}
		
			return BitConverter.ToUInt32 (goodBytes,0);
		}


		/************************************************************
		 * Function: readVarLen()
		 * Purpose: Midis use a weird type of storing numbers to save 
		 *          space. We use this to reconstruct those numbers
		 ************************************************************/
		UInt32 ReadVarLen() 
		{
			UInt32 value;
			byte [] c = new byte [1];
			int cnt = 0;
			
			fs.Read (c,0,1);
			value = Convert.ToUInt32 (c[0]);
			
			/* we dont need variable stuff if its one bit */
			if ((value & 0x80) == 0) 
				return (value);
			
			/* Handle many bits */
			value &= 0x7F;
			do
			{
				fs.Read (c,0,1);
				value = (value << 7) + (Convert.ToUInt32 (c [0]) & 0x7F);
				cnt ++;
			} while (((Convert.ToUInt32 (c [0]) & 0x80) != 0) && (cnt < 4));
			
			return (value);
		}


		/*************************************************************
		 * Function: getVariableString()
		 * purpose: gets a string of variable size from the filestream
		 ************************************************************/
		string GetVariableString() 
		{
			Int32 metalength = Convert.ToInt32 (ReadVarLen ());
			byte[] dataReadB = new byte[metalength];
			fs.Read (dataReadB,0,metalength);
			string dataRead = System.Text.Encoding.ASCII.GetString (dataReadB);
			
			/* Strip the new Lines.. They are annoying */
			dataRead = dataRead.Replace ("\r"," ");
			dataRead = dataRead.Replace ("\n"," ");
			
			return dataRead;		
		}
		
		/*********************************************************
		 * Function: decodeFrame() 
		 * Purpose:  Decodes a single frame in a midi file. Its
		 *           either a header or a track
		 *********************************************************/
		bool DecodeFrame() 
		{
			byte [] trkTypeB 	= new byte [4];
				
			if (fs.Read (trkTypeB,0,4) <= 0) 
				return false;
		
			string trkType 		= System.Text.Encoding.ASCII.GetString (trkTypeB);
			byte [] bTrkLen 	= new byte [4];
			fs.Read (bTrkLen,0,4);
			UInt32 trkLen 		= Convert.ToUInt32 (UInt32ConvEndian (bTrkLen));
			
			/* headers suck.. they have nothing we want */
			if (trkType == "MThd") {
				fs.Seek (trkLen, SeekOrigin.Current);
				return true; 
			}
	
			/* if its not a track, its a corrupt file.. */
			if (trkType != "MTrk") 
				return false;
		
			/* decode the track and return when we reach the end */ 
			while (true) {	
				/* skip the deltatime (variable) */
				ReadVarLen ();
			
				/* Fetch the eventType */
				byte [] eventType = new byte [1];
				fs.Read (eventType,0,1);
			
				/* Pretend the track is ended */
				if (fs.Length == fs.Position) {
					return false;
				}
			
				/* Its a sybex event, skip it */
				if (eventType [0] == 0xF0) {
					UInt32 len = ReadVarLen ();
					fs.Seek (len,SeekOrigin.Current); 
					continue;
				}
				
				/* If its not a meta event, we dont care */
				if (eventType [0] != 0xFF)
					continue;
				
				/* Else its a meta event.Lets get the Meta-event type */
				byte [] metaTypeB= new byte [1];
				fs.Read (metaTypeB,0,1);	
				
				switch(metaTypeB [0]) {
				case 0x2F:
					/* This track is finished.. Skip the \00 and quit the block */ 					
					fs.Seek (1,SeekOrigin.Current);
					return true;
				
				case 0x00: 
					/* Skip The Sequence Number */
					byte [] seqB = new byte [1];
					fs.Read (seqB,0,1);
	
					if (seqB [0] == 0x02)
						fs.Seek (2,SeekOrigin.Current);
					continue;
				
				case 0x01:
					/* extract Text */
					randomText += GetVariableString() + " ";
					continue;
	
				case 0x02: 
					/* extract Copyright */
					copyright += GetVariableString() + " ";
					continue;
	
				case 0x03: 
					/* extract the name of the sequence or track */
					trackName += GetVariableString () + " ";
					continue;
	
				case 0x05: 
					/* extract lyrics */
					lyrics += GetVariableString ();
					continue;
				
				/* Skip the cue points, markers, etc or a proprietry event */
				case 0x04: 
					goto case 0x7F;	
				case 0x06:
					goto case 0x7F;
				case 0x07:
					goto case 0x7F;
				case 0x08:
					goto case 0x7F;
				case 0x09:
					goto case 0x7F;
				case 0x7F:
					fs.Seek (ReadVarLen (),SeekOrigin.Current);
					continue;
	
				case 0x51:
					/* Skip the Tempo */
					fs.Seek (4,SeekOrigin.Current);
					continue;
				
				case 0x54:
					/* Skip the SMPTE offset */
					fs.Seek (6,SeekOrigin.Current);
					continue;
	
				case 0x58:
					/* skip the time signature */
					fs.Seek (5,SeekOrigin.Current);
					continue;
				
				case 0x59:
					/* skip the key signature */
					fs.Seek (3,SeekOrigin.Current);
					continue;
				
				case 0x21: 
					/* Skip the channel prefix */
					fs.Seek (ReadVarLen (),SeekOrigin.Current);
					continue;
					
				case 0x20:
					/* skip the channel port */
					fs.Seek (ReadVarLen (),SeekOrigin.Current);
					continue;
				
				default:
					/* If we get here, skip everything */
					fs.Seek (trkLen - 3, SeekOrigin.Current);	
					//Console.WriteLine ("Bad meta event.Ignoring frame");
					return true;
				}
			} /* End mtrk While(1) */
		}

		/*************************************************************
		 * Function: doPullProperties()
		 * Purpose: Gets the properties of the file
		 ************************************************************/
		protected override void DoPullProperties() 
		{

			/* loop through every frame till the end */
			while ((fs.Position < fs.Length) && (DecodeFrame ()));

			/* Fetch the properties */
			if (randomText != "") 
				AddProperty (Beagle.Property.New ("IncludedText", randomText));
					
			if (copyright != "")
				AddProperty (Beagle.Property.New ("copyright", copyright));
					
			if (trackName != "")
				AddProperty (Beagle.Property.New ("trackname", trackName));
					
			if (lyrics != "")
				AddProperty (Beagle.Property.New ("lyrics", lyrics));	
		}

		/*************************************************************
		* Function: doOpen()
		* Purpose: Needed for our type of stream reading
		*************************************************************/
		override protected void DoOpen (FileInfo info)
		{
			fs = new FileStream (info.FullName, FileMode.Open, FileAccess.Read);
		}
	}
}
